import 'package:json_api_document/json_api_document.dart';
import 'package:json_api_server/src/collection.dart';
import 'package:json_api_server/src/nullable.dart';
import 'package:json_api_server/src/page.dart';
import 'package:json_api_server/src/request_target.dart';
import 'package:json_api_server/src/routing.dart';

/// The Document builder is used by JsonApiServer. It abstracts the process
/// of building response documents and is responsible for such aspects as
///  adding `meta` and `jsonapi` attributes and generating links
class DocumentBuilder {
  final UriSchema _url;

  const DocumentBuilder(this._url);

  /// A document containing a list of errors
  Document errorDocument(Iterable<JsonApiError> errors) =>
      Document.error(errors);

  /// A collection of (primary) resources
  Document<ResourceCollectionData> collectionDocument(
          Collection<Resource> collection, Uri self,
          {Iterable<Resource> included}) =>
      Document(ResourceCollectionData(collection.elements.map(_resourceObject),
          self: Link(self), pagination: _pagination(collection.page, self)));

  /// A collection of related resources
  Document<ResourceCollectionData> relatedCollectionDocument(
          Collection<Resource> collection, Uri self,
          {Iterable<Resource> included}) =>
      Document(ResourceCollectionData(collection.elements.map(_resourceObject),
          self: Link(self), pagination: _pagination(collection.page, self)));

  /// A single (primary) resource
  Document<ResourceData> resourceDocument(Resource resource, Uri self,
          {Iterable<Resource> included}) =>
      Document(
        ResourceData(_resourceObject(resource),
            self: Link(self), included: included?.map(_resourceObject)),
      );

  /// A single related resource
  Document<ResourceData> relatedResourceDocument(Resource resource, Uri self,
          {Iterable<Resource> included}) =>
      Document(ResourceData(_resourceObject(resource),
          included: included?.map(_resourceObject), self: Link(self)));

  /// A to-many relationship
  Document<ToMany> toManyDocument(Iterable<Identifier> collection,
          RelationshipTarget target, Uri self) =>
      Document(ToMany(collection.map(_identifierObject),
          self: Link(self),
          related:
              Link(_url.related(target.type, target.id, target.relationship))));

  /// A to-one relationship
  Document<ToOne> toOneDocument(
          Identifier identifier, RelationshipTarget target, Uri self) =>
      Document(ToOne(nullable(_identifierObject)(identifier),
          self: Link(self),
          related:
              Link(_url.related(target.type, target.id, target.relationship))));

  /// A document containing just a meta member
  Document metaDocument(Map<String, Object> meta) => Document.empty(meta);

  IdentifierObject _identifierObject(Identifier id) =>
      IdentifierObject(id.type, id.id);

  ResourceObject _resourceObject(Resource resource) {
    final relationships = <String, Relationship>{};
    relationships.addAll(resource.toOne.map((k, v) => MapEntry(
        k,
        ToOne(nullable(_identifierObject)(v),
            self: Link(_url.relationship(resource.type, resource.id, k)),
            related: Link(_url.related(resource.type, resource.id, k))))));

    relationships.addAll(resource.toMany.map((k, v) => MapEntry(
        k,
        ToMany(v.map(_identifierObject),
            self: Link(_url.relationship(resource.type, resource.id, k)),
            related: Link(_url.related(resource.type, resource.id, k))))));

    return ResourceObject(resource.type, resource.id,
        attributes: resource.attributes,
        relationships: relationships,
        self: Link(_url.resource(resource.type, resource.id)));
  }

  Pagination _pagination(Page page, Uri self) {
    if (page == null) return Pagination.empty();
    return Pagination.fromLinks(page.map((_) => Link(_.addTo(self))));
  }
}
